#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <fcntl.h>
#include <time.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <dirent.h>
#include <errno.h>

#include "hfo.h"
#include "server.h"

server_t *setup_server(char *host, int port) {

    server_t *server;
    struct sockaddr_in addr;
    
    server = (server_t *) malloc(sizeof(server_t));
    memset(server, 0, sizeof(server_t));
    
    server->numclients = 0;
    
    if ((server->sock = socket(AF_INET, SOCK_STREAM, 0)) == -1) {
        perror("socket()");
        return NULL;
    }
    
    memset(&addr, 0, sizeof(addr));
    addr.sin_family = AF_INET;
    addr.sin_port = htons((uint16_t) port);
    if (host)
        addr.sin_addr.s_addr = inet_addr(host);
    else
        addr.sin_addr.s_addr = htonl(INADDR_ANY);
    
    if (bind(server->sock, (struct sockaddr *) &addr, sizeof(addr)) == -1) {
        perror("bind()");
        close(server->sock);
        return NULL;
    }
    if (listen(server->sock, BACKLOG_LENGTH) == -1) {
        perror("listen()");
        return NULL;
    }

    return server;
}

int get_connections(server_t *server) {
    
    struct sockaddr_in addr;
    socklen_t addrlen;
    struct timeval timeout; 
    fd_set rfds;
    
    if (server->numclients == MAX_CLIENTS-1)
        return -1;
        
    timeout.tv_sec = 0;
    if (server->numclients > 0)
        timeout.tv_usec = 1;
    else
        timeout.tv_usec = 10;

    FD_ZERO(&rfds);
    FD_SET(server->sock, &rfds);
    
    if ((select(server->sock+1, &rfds, 0, 0, &timeout)) > 0) {
        memset(&addr, 0, sizeof(addr));
        addrlen = sizeof(addr);
        /* printf("Got connection\n"); */
        server->client[server->numclients].sock = accept(server->sock, (struct sockaddr *) &addr, &addrlen);
        
        initialize_client(&server->client[server->numclients], resolv_host(&addr));
        server->numclients++;
        
        return 1;
    }
    return 0;
}

char *resolv_host(struct sockaddr_in *sin) {
    
    struct hostent *hent;
    
    hent = gethostbyaddr(&sin->sin_addr.s_addr, sizeof(sin->sin_addr.s_addr), AF_INET);
    
    return (hent ? hent->h_name : inet_ntoa(sin->sin_addr));
    
}

void initialize_client(client_t *client, char *ip) {
    
    client->host = strdup(ip);
    client->fd = -1;
    client->sent = 0;
    client->request = NULL;
    client->file_offset = 0;
    client->total_size = 0;
    memset(client->rbuf, 0, BUFSIZE+1); 
    client->bytes_in_buf = 0;
}

void read_requests(server_t *server) {
    
    int i;
    struct timeval timeout;
    fd_set rfds;
    
    if (server->numclients > 0) {
        timeout.tv_sec = 0;
        timeout.tv_usec = 0;
        FD_ZERO(&rfds);
    
        for (i = 0; i < server->numclients; i++)
            FD_SET(server->client[i].sock, &rfds);

        if ((select(server->client[server->numclients-1].sock+1, &rfds, 0, 0, &timeout)) > 0) {
            for (i = 0; i < server->numclients; i++) {
                /* printf("numclients: %d\n", server->numclients); */
                if (FD_ISSET(server->client[i].sock, &rfds)) {
                    FD_CLR(server->client[i].sock, &rfds);
                    if (!process_request(&server->client[i]))
                        client_disconnect(server, i);
                }
            }

        }
    }
}

bool_t process_request(client_t *client) {
    
    int ret;
    
    /* printf("process_requests()\n"); */

    ret = recv(client->sock,
               client->rbuf + client->bytes_in_buf,
               BUFSIZE - client->bytes_in_buf, 0);
    
    
    if (ret == 0 || ret == -1 || ret == BUFSIZE) {
        /* cliented dropped connection, read error or too large request */
        return FALSE;
    }
    else {
        /* got data from client */
        client->bytes_in_buf += ret;
        if (!parse_client_request(client))
            return FALSE;;
    }
    
    return TRUE;
}

void client_disconnect(server_t *server, int cli_index) {

    client_t *disc_cli = &server->client[cli_index];
    client_t *last_cli = &server->client[server->numclients-1];
    
    shutdown(disc_cli->sock, SHUT_RDWR);
    close(disc_cli->fd);
    free(disc_cli->request);
    free(disc_cli->host);
    
    /* move last client to the position of the disconnected client */
    if (disc_cli != last_cli) {
        disc_cli->sock = last_cli->sock;
        disc_cli->fd = last_cli->fd;
        disc_cli->host = last_cli->host;
        memcpy(&disc_cli->rbuf, &last_cli->rbuf, BUFSIZE);
        disc_cli->bytes_in_buf = last_cli->bytes_in_buf;
        disc_cli->request = last_cli->request;
        disc_cli->total_size = last_cli->total_size;
        disc_cli->start_time = last_cli->start_time;
        disc_cli->sent = last_cli->sent;
        disc_cli->file_offset = last_cli->file_offset;
    }
    server->numclients--;
}

void send_files(server_t *server) {
    
    int i;
    
    /* printf("send_files()\n"); */
    for (i = 0; i < server->numclients; i++) {
        if (server->client[i].request) {
            /* printf("request found\n"); */
            if (is_valid_request(server, i)) {
                switch (file_type(server->client[i].request)) {
                    case IS_ROOT:
                        send_directory_index(server, i, NULL);
                        client_disconnect(server, i);
                        break;
                    case IS_DIRECTORY:
                        /* printf("IS_DIRECTORY\n"); */
                        send_directory_index(server, i, server->client[i].request);
                        client_disconnect(server, i);
                        break;
                    case IS_REGULAR:
                        /* send file */
                        /* printf("IS_REGULAR\n"); */
                        if (!send_file(&server->client[i]))
                            client_disconnect(server, i);
                        break;
                    case IS_INVALID:
                        /* send 404 */
                        send_not_found(server->client[i].sock, server->client[i].request);
                        client_disconnect(server, i);
                        break;
                    case IS_NO_ACCESS:
                        send_forbidden(server->client[i].sock, server->client[i].request);
                        client_disconnect(server, i);
                        break;
                }
            }
            else {
                send_not_found(server->client[i].sock, server->client[i].request);
                client_disconnect(server, i);
            }
            
        }

    }
}

bool_t is_valid_request(server_t *server, int cli_index) {

    struct stat st;
    int i;
    
    if (server->client[cli_index].request[0] == '/' && server->client[cli_index].request[1] == '\0')
        return TRUE;
    
    for (i = 0; i < server->filenum; i++) {
        if ((lstat(server->filebase[i], &st)) != -1) {
            if (S_ISDIR(st.st_mode)) {
                if ((strncmp(server->filebase[i], &server->client[cli_index].request[1], strlen(server->filebase[i]))) == 0) 
                    return TRUE;
            }
            else {
                if ((strcmp(server->filebase[i], &server->client[cli_index].request[1])) == 0)
                    return TRUE;
            }
        }
    }
    return FALSE;
}
    
char file_type(char *request) {
    
    struct stat st;
    int f;
    DIR *dir;
    
    if (request[0] == '/' && request[1] == '\0')
        return IS_ROOT;
    
    if ((stat(&request[1], &st)) != -1) { 
        if (S_ISDIR(st.st_mode)) {
            if ((dir = opendir(&request[1])) == NULL) {
                return IS_NO_ACCESS;
            }
            closedir(dir);
            return IS_DIRECTORY;
        }
        else if (S_ISREG(st.st_mode)) {
            if ((f = open(&request[1], O_RDONLY)) == -1) {
                return IS_NO_ACCESS;
            }
            close(f);
            return IS_REGULAR;
        }
    }
    
    return IS_INVALID;
}

bool_t send_file(client_t *client) {
    
    char buf[BUFSIZE];
    ssize_t bytes;
    struct stat st;
    struct timeval timeout; 
    fd_set wfds;
    
    if (client->fd == -1) {
        /* no previous file open */
        lstat(&client->request[1], &st);
        client->fd = open(&client->request[1], O_RDONLY); 
        client->total_size = st.st_size;
        client->start_time = time(NULL);
        if (client->file_offset > client->total_size)
            return FALSE;
                
        if (client->file_offset > 0) {
            lseek(client->fd, client->file_offset, SEEK_SET);
            client->total_size -= client->file_offset;
            if (!send_response(client->sock, 206))
                return FALSE;
            if (!send_content_type(client->sock, OCTET_STREAM))
                return FALSE;
            if (!send_content_length(client->sock, client->total_size - client->file_offset))
                return FALSE;
            if (!send_partial_content(client->sock, client->file_offset, client->total_size, client->total_size))
                return FALSE;
            if (!send_end_response(client->sock))
                return FALSE;
        }
        else {
            if (!send_response(client->sock, 200))
                return FALSE;
            if (!send_content_type(client->sock, OCTET_STREAM))
                return FALSE;
            if (!send_content_length(client->sock, client->total_size))
                return FALSE;
            if (!send_end_response(client->sock))
                return FALSE;
        }
        
    }
    
    timeout.tv_sec = 0;
    timeout.tv_usec = 1;
    FD_ZERO(&wfds);
    FD_SET(client->sock, &wfds);
    /* select() to be sure that the send() won't block */
    if ((select(client->sock + 1, 0, &wfds, 0, &timeout)) > 0) {
        /* read data from file */
        bytes = read(client->fd, buf, BUFSIZE);
        if (bytes == 0 || bytes == -1) {
            /* end of file or error reading, disconnect client */
            close(client->fd);
            return FALSE;
        }
        /* send data to socket */ 
        if ((send(client->sock, buf, bytes, 0)) < 0) 
            return FALSE;
        
        client->sent += bytes;
    }
    return TRUE;
    
}

bool_t send_response(int fd, int status_code) {
    
    switch (status_code) {
        case 200:
            if (send(fd, CODE_200_OK, strlen(CODE_200_OK), 0) < 0)
                return FALSE;
            return TRUE;
        case 206:
            if (send(fd, CODE_206_PARTIAL_CONTENT, strlen(CODE_206_PARTIAL_CONTENT), 0) < 0)
                return FALSE;
            return TRUE;
        case 404:
            if (send(fd, CODE_404_NOT_FOUND, strlen(CODE_404_NOT_FOUND), 0) < 0)
                return FALSE;
            return TRUE;
        case 403:
            if (send(fd, CODE_403_FORBIDDEN, strlen(CODE_403_FORBIDDEN), 0) < 0)
                return FALSE;
            return TRUE;
    }
    return FALSE;
}

bool_t send_content_type(int fd, char type) {

    switch (type) {
        case TEXT_HTML:
            if (send(fd, TEXT_HTML_CONTENT, strlen(TEXT_HTML_CONTENT), 0) < 0)
                return FALSE;
            return TRUE;
        case OCTET_STREAM:
            if (send(fd, OCTET_STREAM_CONTENT, strlen(OCTET_STREAM_CONTENT), 0) < 0)
                return FALSE;
            return TRUE;
    }
    return FALSE;
}

bool_t send_end_response(int fd) {
    
    if (send(fd, END_RESPONSE, strlen(END_RESPONSE), 0) < 0)
        return FALSE;

    return TRUE;
}

bool_t send_content_length(int fd, ssize_t len) {
    
    char *resp = (char *) malloc(FILESIZE_MAX_LEN + strlen(CONTENT_LENGTH));
    
    sprintf(resp, CONTENT_LENGTH, len);
    if (send(fd, resp, strlen(resp), 0) < 0) {
        free(resp);
        return FALSE;
    }
    
    free(resp);
    return TRUE;
}

bool_t send_partial_content(int fd, ssize_t start, ssize_t stop, ssize_t size) {

    char *resp = (char *) malloc(3 * FILESIZE_MAX_LEN + strlen(CONTENT_RANGE));
    
    sprintf(resp, CONTENT_RANGE, start, stop, size);
    if (send(fd, resp, strlen(resp), 0) < 0) {
        free(resp);
        return FALSE;
    }
    
    free(resp);
    return TRUE;
    
}

void send_not_found(int fd, char *item) {
    /* -1 because the string contains a %s that will be replaced with 'item' */
    char *resp = (char *) malloc(strlen(item) + strlen(NOT_FOUND_BODY) - 1);
    
    send_response(fd, 404);
    send_content_type(fd, TEXT_HTML);
    send_end_response(fd);
    
    sprintf(resp, NOT_FOUND_BODY, item);
    send(fd, resp, strlen(resp), 0);
    free(resp);
}

void send_forbidden(int fd, char *item) {
    /* -1 because the string contains a %s that will be replaced with 'item' */
    char *resp = (char *) malloc(strlen(item) + strlen(FORBIDDEN_BODY) - 1);
    
    send_response(fd, 403);
    send_content_type(fd, TEXT_HTML);
    send_end_response(fd);
    
    sprintf(resp, FORBIDDEN_BODY, item);
    send(fd, resp, strlen(resp), 0);
    free(resp);
}

bool_t parse_client_request(client_t *client) {
    
    char *end;
    char *range;
    
    /* check if the request is complete */
    if (strstr(client->rbuf, CRLF CRLF)) { 
        /* check for a requested range */
        if ((range = strstr(client->rbuf, RANGE_REQUEST))) 
            client->file_offset = atoi(range+strlen(RANGE_REQUEST));

        /* look for the end of the first line */
        for (end = client->rbuf; (*end != '\r') && (*end != '\0'); end++)
            ;

        if (*end != '\r' || !IS_GET_REQUEST)
            /* not a valid request, only GET is supported */
            return FALSE;

        /* remove the trailing HTTP/x.x */
        while (*end != ' ')
            end--;

        /* we want to get rid of the ' ' too */
        *end = '\0'; 
        /* save the request without GET infront */
        client->request = strdup(&client->rbuf[4]);    

        /* printf("GET REQUEST: '%s'\n", client->request); */
        return TRUE;
    }
    
    return TRUE;
}

bool_t send_directory_index(server_t *server, int cli_index, char *directory) {

    int i;
    DIR *dir;
    struct dirent *dent;
    char *entry;
    char *path;
    
    /* printf("send_directory_index()\n"); */
    
    if (!send_response(server->client[cli_index].sock, 200))
        return FALSE;
    if (!send_content_type(server->client[cli_index].sock, TEXT_HTML))
        return FALSE;
    if (!send_end_response(server->client[cli_index].sock))
        return FALSE;
            
    if ((send(server->client[cli_index].sock, INDEX_HEADER, strlen(INDEX_HEADER), 0)) <= 0)
        return FALSE;
    
    if (!directory) {
        /* printf("after if dir\nserver->filenum: %d\n", server->filenum);  */
        for (i = 0; i < server->filenum; i++) {
            if ((entry = create_file_entry(server->filebase[i])) != NULL) {
                if ((send(server->client[cli_index].sock, entry, strlen(entry), 0)) <= 0) {
                    free(entry);
                    return FALSE;
                }
                free(entry);
            }
            /* FIXME: needs to be free():d even if send is unsuccessful */
        }
    }
    else {
        
        dir = opendir(&directory[1]);
                
        while ((dent = readdir(dir)) != NULL) {
            /* ignore entrys starting with a . */
            if (dent->d_name[0] != '.') {
                path = (char *) malloc(strlen(&directory[1])+strlen(dent->d_name)+1);
                memset(path, 0, sizeof(path));
                strcpy(path, &directory[1]);
                strcat(path, dent->d_name);
                if ((entry = create_file_entry(path)) != NULL) {
                    if ((send(server->client[cli_index].sock, entry, strlen(entry), 0)) <= 0) {
                        free(entry);
                        free(path);
                        return FALSE;
                    }
                    free(entry);
                }
                /* FIXME: needs to be free():d even if send is unsuccessful */
                free(path);
            }
        }
        closedir(dir);
    }
    
    if ((send(server->client[cli_index].sock, INDEX_FOOTER, strlen(INDEX_FOOTER), 0)) <= 0)
        return FALSE;
    
    return TRUE;
}

char *base_filename(char *filename) {
    
    char *p;
    if ((p = strrchr(filename, '/')) != NULL) 
        return ++p;
    
    return filename;
}

char *create_file_entry(char *file) {
    
    struct stat st;
    char *entry;
    char size[FILESIZE_MAX_LEN];
    
    if ((lstat(file, &st)) != -1) {
        if (S_ISDIR(st.st_mode)) 
            sprintf(size, "%s", "-"); 
        else if (S_ISREG(st.st_mode)) 
            sprintf(size, "%d", (int) st.st_size);
        
        entry = (char *) malloc(strlen(file)+
                1+
                strlen(base_filename(file))+
                NUM_SPACE_BEFORE_SIZE+
                strlen(INDEX_ENTRY_START)+
                strlen(INDEX_ENTRY_MIDDLE)+
                strlen(INDEX_ENTRY_END)+
                strlen(size)+
                strlen(CRLF)+
                (S_ISDIR(st.st_mode) ? 2 : 0) + 
                1);
        
        sprintf(entry, "%s/%s%s%s%s%s%s%-*.*s%s%s", INDEX_ENTRY_START, 
                file,
                S_ISDIR(st.st_mode) ? "/" : "",
                INDEX_ENTRY_MIDDLE,
                base_filename(file),
                S_ISDIR(st.st_mode) ? "/" : "",
                INDEX_ENTRY_END,
                NUM_SPACE_BEFORE_SIZE,
                NUM_SPACE_BEFORE_SIZE,
                " ",
                size,
                CRLF);
      
        return entry;
    } 
    
    return NULL;
}

server_t *shutdown_server(server_t *server) {

    int i;

    for (i = 0; i < server->numclients; i++) {
        free(server->client[i].host);
        free(server->client[i].request);
        shutdown(server->client[i].sock, SHUT_RDWR);
        close(server->client[i].fd);
    }
    
    free(server->filebase);
    shutdown(server->sock, SHUT_RDWR);
    free(server);
    
    return NULL;
}
